import random

from lunas import Iterator, TextLine, Zip, Shuffle


def generate():
    with open('src', 'w') as w:
        for i in range(100000):
            w.write(f'{i}.{random.randint(0, 100000)}\n')

    with open('trg', 'w') as w:
        for i in range(100000):
            w.write(f'{i}.{random.randint(0, 100000)}\n')


def clean():
    import os
    os.remove('src')
    os.remove('trg')


def aggregate_value_by_key(dicts, key, reduction=None):
    values = [d[key] for d in dicts]
    if reduction:
        return reduction(values)
    return values


def get_it():
    threads = 6
    src = TextLine('src', num_threads=threads)
    trg = TextLine('trg', num_threads=threads)
    ds = Zip([src, trg], num_threads=threads)

    ds = ds.select(lambda x, y: {
        'xi': int(x.split('.')[0]),
        'yi': int(y.split('.')[0]),
        'x': int(x.split('.')[1]),
        'y': int(y.split('.')[1]),
        'X': x,
        'Y': y,
    })

    shuffle = -1
    if shuffle != 0:
        ds = Shuffle(ds, shufsize=shuffle, num_threads=threads)

    ds = ds.where(
        lambda x: x['x'] % 2 == 0 and x['y'] % 2 == 0
    )

    def collate_fn(xs):
        return {
            'xi': aggregate_value_by_key(xs, 'xi'),
            'yi': aggregate_value_by_key(xs, 'yi'),
            'x': aggregate_value_by_key(xs, 'x'),
            'y': aggregate_value_by_key(xs, 'y'),
            'X': aggregate_value_by_key(xs, 'X'),
            'Y': aggregate_value_by_key(xs, 'Y'),
        }

    sample_size_fn = None

    batch_size = 10

    iterator = Iterator(
        ds, batch_size,
        cache_size=max(20, 1) * batch_size,
        sample_size_fn=sample_size_fn,
        collate_fn=collate_fn,
        sort_cache_by=lambda sample: -sample['y'],
        sort_batch_by=lambda sample: -sample['x'],
        strip_batch=True
    )

    return iterator


def test_align():
    it = get_it()
    for batch in it.iter_epoch():
        assert batch.data['xi'] == batch.data['yi']


def test_filter():
    it = get_it()
    import numpy
    for batch in it.iter_epoch():
        x = numpy.array(batch.data['x'])
        y = numpy.array(batch.data['y'])
        assert numpy.all(x % 2 == 0)
        assert numpy.all(y % 2 == 0)


generate()
test_align()
test_filter()
clean()



import itertools
from multiprocessing.pool import ThreadPool
from typing import Callable, Any, List


def parallel_map(fn: Callable[[Any], Any], inputs: List[Any], num_thread: int):
    """Applies a function to a list of inputs in parallel.

    Uses a Threading pool to process inputs in parallel.

    Args:
        fn: A `Callable` function that takes an element from `inputs` as argument.
        inputs: A `Iterable` object.
        num_thread: An `int` scalar represents the number of threads.

    Returns:
        A corresponding list of processed inputs.
    """
    if num_thread <= 1:
        return list(map(fn, inputs))
    pool = ThreadPool(num_thread)
    results = pool.map(fn, inputs)

    pool.close()
    pool.join()
    return results


def try_get_attr(instance, attr_name):
    if hasattr(instance, attr_name):
        return getattr(instance, attr_name)
    else:
        return None


def get_state_dict(obj, recursive=True, exclusions=None, inclusions=None):
    keys = vars(obj).keys()
    state_dict = {}
    if exclusions and inclusions:
        raise TypeError('Argument `exclusions` and `inclusions` '
                        'should not be both non-empty at the same time.')

    if exclusions:
        keys = [k for k in keys if k not in exclusions]
    elif inclusions:
        keys = [k for k in keys if k in inclusions]

    props = map(lambda k: getattr(obj, k), keys)

    for key, prop in itertools.zip_longest(keys, props):
        if recursive:
            to_state_dict = try_get_attr(prop, 'state_dict')
            if callable(to_state_dict):
                prop = to_state_dict()
        state_dict[key] = prop

    return state_dict


def load_state_dict(obj, state_dict):
    for key, val in state_dict.items():
        prop = getattr(obj, key)
        load_state_dict_ = try_get_attr(prop, 'load_state_dict')
        if callable(load_state_dict_):
            load_state_dict_(val)
        else:
            setattr(obj, key, val)



import sys

from overrides import overrides

from lunas.readers.base import BaseReader


class Stdin(BaseReader):
    def __init__(self, bufsize: int = 10000, num_threads: int = 1,sentinel=''):
        super().__init__(bufsize, num_threads)
        self._iterator = None
        self._sentinel=sentinel
        self._num_line = 0

    @overrides
    def size(self) -> int:
        return sys.maxsize

    @overrides
    def next(self):
        line=next(self._iterator)
        self._num_line += 1
        return line

    @overrides
    def _reset_cursor(self):
        self._reset()
        super()._reset_cursor()

    def _reset(self):
        self._iterator = iter(sys.stdin.readline, self._sentinel)

    @overrides
    def _finalize(self):
        self._num_line = 0
        self._iterator = None
        super()._finalize()



import itertools
from collections import deque
from typing import List, Dict

import numpy
from overrides import overrides

from lunas.readers.base import BaseReader


class Nested(BaseReader):
    def __init__(self, reader, bufsize: int = 10000, num_threads: int = 1):
        super().__init__(bufsize, num_threads)
        multi_reader = isinstance(reader, (list, tuple))
        if not multi_reader:
            reader = [reader]
        for r in reader:
            r._fast_skip = True
        self.reader = reader
        self._multi_reader = multi_reader
        self._inclusions += ['reader', '_multi_reader']

    @overrides
    def size(self) -> int:
        return len(self.reader[0])

    @overrides
    def next(self):
        sample = tuple([r._buffered_next() for r in self.reader])
        return sample if self._multi_reader else sample[0]

    @overrides
    def state_dict(self) -> Dict:
        state = super().state_dict()
        state['reader'] = [r.state_dict() for r in self.reader]
        return state

    @overrides
    def load_state_dict(self, state_dict: Dict) -> None:
        for r, state in zip(self.reader, state_dict['reader']):
            r.load_state_dict(state)
        del state_dict['reader']

        super().load_state_dict(state_dict)

    @overrides
    def _process_buffer(self, buffer=None, inplace=True):
        buffer = self._buffer if buffer is None else buffer
        reader = self.reader
        if self._multi_reader:
            new_buffer = list(zip(
                *[r._process_buffer(buf, inplace=False) for r, buf in
                  zip(reader, itertools.zip_longest(*buffer))]))
            new_buffer = list(map(
                lambda sample: sample if all(s is not None for s in sample) else None,
                new_buffer
            ))
        else:
            new_buffer = reader[0]._process_buffer(buffer, inplace=False)

        return super()._process_buffer(new_buffer, inplace)

    @overrides
    def _finalize(self):
        for reader in self.reader:
            reader._finalize()
        super()._finalize()

    @overrides
    def __iter__(self):
        self.reader = list(map(iter, self.reader))
        return super().__iter__()


class Shuffle(Nested):
    def __init__(self, reader: BaseReader, shufsize=None, bufsize: int = 10000, num_threads: int = 1):
        """
        Apply random permutation to a given reader.
        Args:
            reader:
            shufsize: sequentially read samples from a reader into a fixed-sized buffer for permutation.
            bufsize: process samples in buffer in parallel.
            num_threads: number of threads for parallel processing.
        """
        if not shufsize or shufsize < 0:
            shufsize = len(reader)
        shufsize = shufsize or bufsize

        assert shufsize > 0, shufsize

        super().__init__(reader, bufsize, num_threads)

        self._shufsize = shufsize
        self._shuffled_buffer: deque = deque()
        self._random_state = numpy.random.get_state()

        self._inclusions += ['_random_state']

    @overrides
    def next(self):
        if not self._shuffled_buffer:
            self._shuffle()
        try:
            sample = self._shuffled_buffer.popleft()
        except IndexError:
            raise StopIteration
        return sample

    @overrides
    def load_state_dict(self, state_dict: Dict) -> None:
        numpy.random.set_state(state_dict['_random_state'])
        del state_dict['_random_state']
        super().load_state_dict(state_dict)

    def _shuffle(self):
        buffer = self._shuffled_buffer
        while len(buffer) < self._shufsize:
            try:
                buffer.append(super().next())
            except StopIteration:
                break
        buffer = list(buffer)
        numpy.random.shuffle(buffer)
        self._shuffled_buffer = deque(buffer)

    @overrides
    def __iter__(self):
        self._random_state = numpy.random.get_state()
        return super().__iter__()


class Zip(Nested):
    def __init__(self, reader: List[BaseReader], bufsize: int = 10000, num_threads: int = 1):
        super().__init__(reader, bufsize, num_threads)
        sizes = list(map(len, reader))
        if len(set(sizes)) != 1:
            raise RuntimeError(
                f'Sizes of datasets {tuple(sizes)} must match.'
            )
        self._inclusions += ['reader']



from collections import deque
from typing import List, Dict, Callable, Any

from overrides import overrides

from lunas.batch import Batch, Cache
from lunas.persistable import Persistable
from lunas.readers import BaseReader
from lunas.utils import get_state_dict, load_state_dict


class Iterator(Persistable):
    """An iterator that iterates through a `Reader`.

    This class performs multi-pass iterations over the dataset and maintains
    the iteration state.
    """

    def __init__(self, reader: BaseReader, batch_size, cache_size: int = 1000, sample_size_fn: Callable[[Any], int] = None,
                 collate_fn: Callable[[List[Any]], Any] = None, sort_cache_by: Callable[[Any], int] = None,
                 sort_batch_by: Callable[[Any], int] = None,
                 drop_tails=False, strip_batch=False):
        """Initialize the iterator.

        Args:
            reader: A `Reader` object.
            batch_size: A `int` scalar that limits the size of returned batch.
            cache_size: A `int` scalar. Prefetch `cache_size` samples from the `reader` in `self.cache`.
            sample_size_fn: (Optional.) A callable function that calculates size for each sample.
                The size of each sample will then be summed up as the size of the batch. If not
                specified, default to 1 for each sample, which is equivalent to `lambda sample: 1`.
            collate_fn: (Optional.) A callable function that converts a list of samples to model inputs.
            sort_cache_by: (Optional.) A callable function that returns a sorting key for each sample. If not
                specified, leave the cache as it is. The samples will be sorted in ascending order.
            sort_batch_by: (Optional.) A callable function that returns a sorting key for each sample. If not
                specified, leave the batch as it is. The samples will be sorted in ascending order.
            drop_tails: (Optional.) Whether the last samples of the dataset that cannot fill a batch should be dropped.
            strip_batch:
        """
        super().__init__()
        self._reader = reader
        self._batch_size = batch_size
        self._sample_size_fn = sample_size_fn
        self._collate_fn = collate_fn
        self._cache_size = cache_size
        self._sort_cache_by = sort_cache_by
        self._sort_batch_by = sort_batch_by
        self._drop_tails = drop_tails
        self._strip_batch = strip_batch

        self._sample_size_fn = sample_size_fn
        self._collate_fn = collate_fn

        # bookkeeping params
        self._step_in_epoch = 0
        self._step = 0
        self._epoch = 0

        self._cache = Cache(cache_size, sample_size_fn)
        self._remains: deque = deque()
        self._stripped: deque = deque()

        self._inclusions = ['_inclusions', '_reader', '_step', '_step_in_epoch', '_epoch', '_cache', '_remains',
                            '_stripped']

        self.check_batch_size(batch_size, cache_size)
        self.reset()

    @property
    def cache_size(self):
        return self._cache_size

    @property
    def step_in_epoch(self):
        return self._step_in_epoch

    @property
    def step(self):
        return self._step

    @property
    def epoch(self):
        return self._epoch

    @property
    def batch_size(self):
        return self._batch_size

    def set_batch_size(self, batch_size) -> None:
        """Allows dynamic batch size at runtime.

        Args:
            batch_size: A `int` scalar.

        """
        self.check_batch_size(batch_size)
        self._batch_size = batch_size

    def check_batch_size(self, batch_size, cache_size=None) -> None:
        """Checks whether batch_size is < cache_size.

        To ensure rationality, batch_size must be < cache_size.

        Args:
            batch_size: A `int` scalar.
            cache_size: A `int` scalar.

        """
        cache_size = cache_size or self._cache_size
        if batch_size > cache_size:
            raise RuntimeError(
                f'Batch size ({batch_size}) should be less than cache size ({cache_size}). '
                f'Please lower the batch size or increase the cache size.'
            )

    def reset(self):
        self._step_in_epoch = 0
        self._step = 0
        self._epoch = 0
        self._remains.clear()
        self._cache.pop_all()  # discard
        self._reader = iter(self._reader)

    def reset_epoch(self):
        self._step_in_epoch = 0
        self._remains.clear()
        self._cache.pop_all()

    def iter_epoch(self, before_epoch=None, after_epoch=None):
        """Iterate through the dataset for one epoch.

        For the last batch, it will be dropped if its size is smaller
        than 2/3 of the specified batch size.

        """
        # self.reset_epoch()
        cache = self._cache
        remains = self._remains
        stripped = self._stripped

        end_of_epoch = False

        sort_batch = False
        if before_epoch is not None and self.step_in_epoch == 0:
            before_epoch()

        while True:
            batch = Batch(self.batch_size, self._sample_size_fn)
            if cache.effective_size < self.batch_size * 2 / 3.0:
                if end_of_epoch:
                    # Raise error when the whole dataset cannot form a batch
                    if self.step == 0:
                        raise RuntimeError(
                            f'Size of the dataset ({len(remains)}) '
                            f'is smaller than batch size ({self.batch_size}). '
                            f'Please lower the batch size or '
                            f'check whether the dataset is too small.'
                        )
                    self._reader = iter(self._reader)

                    if self._drop_tails or len(remains) == 0:
                        break
                    else:
                        # The last batch
                        batch.from_deque(remains, self.batch_size)
                        batch.from_iter(cache, self.batch_size)
                        batch.sort(self._sort_batch_by or self._sort_cache_by)
                        self._step_in_epoch += 1
                        self._step += 1
                        yield self._prepare_batch(batch)
                        break

                # Consume samples from cache before filling-in
                remains += cache.pop_all()
                try:
                    # Fill cache
                    cache.from_iter(self._reader, raise_when_stopped=True)
                except StopIteration:
                    # Mark as end
                    end_of_epoch = True
                cache.sort(self._sort_cache_by)

            if self.batch_size == self.cache_size:
                # Simply return the cache as a batch to avoid sorting again.
                batch = cache
                cache = Cache(self.cache_size, self._sample_size_fn)
                self._cache = cache
            else:
                if stripped:
                    batch.from_deque(stripped, self.batch_size)
                if remains:
                    batch.from_deque(remains, self.batch_size)
                    sort_batch = True
                size_diff = batch.from_iter(cache, self.batch_size)
                sort_batch = size_diff > 0 or sort_batch

            if batch.effective_size < self.batch_size:
                # the cache is exhausted while batch is not filled
                # for the last unfilled batch, revert it
                if end_of_epoch:
                    batch.revert()
                remains += batch.pop_all()
            else:
                # filled
                # strip batch size
                if self._strip_batch:
                    stripped += batch.strip(self.batch_size)

                if sort_batch:
                    batch.sort(self._sort_batch_by or self._sort_cache_by)
                    sort_batch = False

                self._step_in_epoch += 1
                self._step += 1
                yield self._prepare_batch(batch)

        if after_epoch is not None:
            after_epoch()

        self._epoch += 1
        self._step_in_epoch = 0

    def while_true(self, predicate: Callable[[], bool], before_epoch=None, after_epoch=None):
        """Iterates through the dataset by a given stopping criteria.

        Args:
            predicate: A callable function. This function is evaluated to determine
            whether iteration should continue or not.
            before_epoch:
            after_epoch:

        Returns:
            (batch, inputs): A `Tuple` consists of a `Batch` object and model inputs. When `self.collate_fn`
                is None, the returned `inputs` is also None.
        """
        epoch_iter = self.iter_epoch(before_epoch, after_epoch)

        if predicate is not None:
            while predicate():
                try:
                    batch = next(epoch_iter)
                except StopIteration:
                    epoch_iter = self.iter_epoch(before_epoch, after_epoch)
                    continue

                yield batch
        else:
            for batch in epoch_iter:
                yield batch

    @overrides
    def state_dict(self) -> Dict:
        return get_state_dict(self, recursive=True, inclusions=self._inclusions)

    @overrides
    def load_state_dict(self, state_dict: Dict) -> None:
        load_state_dict(self, state_dict)

    def _prepare_batch(self, batch: Batch):
        if self._collate_fn:
            batch.process(self._collate_fn)
        return batch

    def __call__(self, while_predicate: Callable[[], bool] = None, before_epoch=None, after_epoch=None):
        return self.while_true(while_predicate, before_epoch, after_epoch)



from overrides import overrides

from lunas.readers.base import BaseReader


class Range(BaseReader):
    def __init__(self, start: int, stop: int = None, step: int = None, bufsize: int = 10000, num_threads: int = 1):
        super().__init__(bufsize, num_threads)
        if stop is None:
            stop = start
            start = 0
        if step is None:
            step = 1

        self._start = start
        self._stop = stop
        self._step = step

        self._range = None

        self._inclusions += ['_range']

    @overrides
    def size(self) -> int:
        import math
        return int(math.ceil((self._stop - self._start) / self._step))

    @overrides
    def next(self):
        return next(self._range)

    def _reset(self):
        start, stop, step = self._start, self._stop, self._step
        self._range = iter(range(start, stop, step))

    @overrides
    def _reset_cursor(self):
        super()._reset_cursor()
        self._reset()



from __future__ import annotations

import abc
from collections import deque
from typing import List, Dict, Callable, Any

from lunas.persistable import Persistable
from lunas.utils import parallel_map, get_state_dict, load_state_dict


class BaseReader(Persistable):
    """A abstract class representing a dataset reader.

    A reader includes pipeline to process each sample and iterate through the dataset.
    This class defines any interface that's visible to users.
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, bufsize: int = 10000, num_threads: int = 1):
        super().__init__()
        self._bufsize = bufsize
        self._num_threads = num_threads
        self._buffer: deque = deque()
        self._fns: List = []
        self._fast_skip: bool = False
        # Indicates position in the dataset.
        self._cursor: int = -1
        self._stop_iteration = False

        # Excludes these attributes from `self.state_dict()`
        self._inclusions: List[str] = ['_inclusions', '_fast_skip', '_cursor', '_stop_iteration']

    def size(self) -> int:
        """Get size of this dataset.

        Note: The returned value represents the total number of samples of the dataset, without filters
        applied to samples. When filters are applied, the effective size of dataset may be different
        from the value returned by this method. So please make sure the filters are applied to the final dataset
        wrapper to avoid size inconsistency.

        For example, the following usage of `self.where()` filters would cause size inconsistency.
        ```python
        ds1 = Range(10).where(lambda x: x<5)
        ds2 = Range(10).where(lambda x: x<6)
        ds = Zip(ds1, ds2)
        iterator = DataIterator(ds)
        # ...
        for batch in iterator():
            print(batch)
        ```

        The effective size of `ds1` at runtime is 5, while it's 6 for `ds2`. This inconsistency cannot be checked
        before iterating through `ds`.

        Returns:
            A scalar of type `int`.

        """

        raise NotImplementedError

    def next(self) -> Any:
        """Get an original sample from the dataset by index. DO NOT invoke this method from
        outside, use `next(dataset)` instead. This method should raise StopIteration at the end of dataset.

        Returns:
            A sample of any type.
        """
        raise NotImplementedError

    def select(self, fn: Callable[[Any], Any]) -> BaseReader:
        """Transforms a sample.

        A fn is a callable object that takes a sample as input and returns a transformed
        sample.

        Note that the transformation is not applied to the dataset immediately. Instead, we
        apply transformations to a large buffer and utilizes multi-threading pool to speed up data
        pipeline.

        Args:
            fn: A `Callable` object.

        Returns:
            `self`.

        """
        self._fns.append(fn)
        return self

    def where(self, predicate: Callable[[Any], bool]) -> BaseReader:
        """Filters a sample by predicate.

        A predicate is a callable object that takes a sample as input and returns a
        `bool` value to indicate whether this sample should be filtered out.

        Note that the predicate is not applied to the dataset immediately. Instead, we
        apply predicates to a large buffer and utilizes multi-threading pool to speed up data
        pipeline.

        Args:
            predicate: A `Callable` object.

        Returns:
            `self`.
        """
        self._fns.append(predicate)
        return self

    def state_dict(self) -> Dict:
        return get_state_dict(self, inclusions=self._inclusions)

    def load_state_dict(self, state_dict: Dict) -> None:
        self._buffer.clear()
        load_state_dict(self, state_dict)

        cursor = self._cursor
        self._reset_cursor()
        # Fast-skip these samples
        is_fast_skip = self._fast_skip
        self._fast_skip = True

        while self._cursor < cursor:
            self._move_cursor()
            self._buffered_next()
        self._fast_skip = is_fast_skip
        if not is_fast_skip:
            self._process_buffer()

    def _move_cursor(self) -> int:
        """Increases the cursor and returns the new value.

        Returns:
            A `int` indicates current position.
        """
        self._cursor += 1
        return self._cursor

    def _reset_cursor(self):
        """Reset cursor to enable re-iterating over the dataset.

        """
        self._stop_iteration = False
        self._cursor: int = -1

    def _fill_buffer(self) -> int:
        size = 0
        if not self._stop_iteration:
            buffer = self._buffer
            for _ in range(self._bufsize - len(buffer)):
                try:
                    buffer.append(self.next())
                except StopIteration:
                    self._stop_iteration = True
                    break
            size = len(buffer)
            if not self._fast_skip and size > 0:
                self._process_buffer()
        return size

    def _process_buffer(self, buffer=None, inplace=True):
        buffer = self._buffer if buffer is None else buffer
        buffer = self._parallel_apply(buffer)
        # buffer = list(filter(None, buffer))

        if inplace:
            self._buffer = deque(buffer)
        return buffer

    def _buffered_next(self) -> Any:
        """Wrapper method to get next sample.

        Wraps `self.next()`, apply transformations and predicates, and maintains
        internal state of this object.

        Returns:
            A sample of any type.

        """
        try:
            return self._buffer.popleft()
        except IndexError:
            size = self._fill_buffer()
            if size == 0:
                raise StopIteration
            return self._buffer.popleft()

    def _apply(self, sample: Any) -> Any:
        """Applies transformations and filters to a sample.

        When sample is filtered out, returns None, else returns transformed sample.

        Args:
            sample: An input sample to be processed.

        Returns:
            A sample or `None` if it's filtered.
        """
        if sample is None:
            return None

        for fn in self._fns:
            new_sample = fn(sample)

            # Stops when predicate is evaluated to False.
            if new_sample is False:
                return None
            elif new_sample is not True:
                sample = new_sample

        return sample

    def _star_apply(self, sample: Any) -> Any:
        if sample is None:
            return None

        for fn in self._fns:
            try:
                new_sample = fn(*sample)
            except TypeError as e:
                raise Exception(f'When a reader returns a tuple or list as a sample, '
                                f'the mapping functions (`Reader.select()` and `Reader.where()`) '
                                f'for succeeding readers should take equivalent number of parameters as args. '
                                f'Please check the mapping functions for `{type(self)}`') from e

            # Stops when predicate is evaluated to False.
            if new_sample is False:
                return None
            elif new_sample is not True:
                sample = new_sample

        return sample

    def _parallel_apply(self, samples: List[Any]) -> List[Any]:
        if self._fns and samples:
            if not isinstance(samples, (tuple, list)):
                samples = list(samples)
            if isinstance(samples[0], (tuple, list)):
                return parallel_map(self._star_apply, samples, self._num_threads)
            else:
                return parallel_map(self._apply, samples, self._num_threads)
        else:
            return samples

    def _finalize(self):
        """Release resources after iteration stops.

        """
        pass

    def __next__(self):
        """Implements iterator method and returns a sample.

        Returns:
            A sample instance.
        """
        try:
            sample = self._buffered_next()
            self._move_cursor()
        except StopIteration as e:
            self._finalize()
            raise e

        if sample is None:
            return self.__next__()
        return sample

    def __iter__(self) -> BaseReader:
        """Returns an iterator of this dataset.

        """
        self._reset_cursor()
        return self

    def __len__(self):
        """Returns size of this dataset.

        Returns:
            A `int` scalar.
        """
        return self.size()



from lunas.readers.base import BaseReader
from lunas.readers.nested import Zip, Shuffle
from lunas.readers.range import Range
from lunas.readers.stdin import Stdin
from lunas.readers.text import TextLine

__all__ = ['BaseReader', 'Zip', 'Shuffle', 'TextLine', 'Range', 'Stdin']



from collections import deque
from typing import List, Iterable, Dict, Callable, Any

import numpy
from overrides import overrides

from lunas.persistable import Persistable
from lunas.utils import get_state_dict, load_state_dict


class Batch(Persistable):
    def __init__(self, max_size: int, size_fn: Callable[[Any], int] = None):
        super().__init__()
        self._max_size = max_size
        self._size_fn = size_fn
        self._samples: deque = deque()
        self._effective_size: int = 0
        self._sort_idx: numpy.ndarray = None
        self._exclusions = ['_size_fn']
        self._data = None

    @property
    def data(self):
        return self._data

    @property
    def samples(self):
        return [sample for sample, size in self._samples]

    @property
    def sizes(self):
        return [size for sample, size in self._samples]

    @property
    def effective_size(self):
        return self._effective_size

    def pop_all(self):
        samples = self.samples
        self._samples.clear()
        self._effective_size = 0
        return samples

    def pop(self, fifo=True):
        if fifo:
            sample, size = self._samples.popleft()
        else:
            sample, size = self._samples.pop()
        self._effective_size -= size
        return sample

    def push(self, sample):
        size = 1
        if self._size_fn is not None:
            size = self._size_fn(sample)
        sample = (sample, size)
        self._samples.append(sample)
        self._effective_size += size

    def strip(self, max_size=None):
        max_size = max_size or self._max_size
        rv = []
        while self.effective_size > max_size:
            rv.append(self.pop(False))

        return rv[::-1]

    def from_iter(self, sample_iter, size: int = None, raise_when_stopped: bool = False):
        """
        Fills batch from an iterable object. Enables dynamic batch size if size is not None.
        Args:
            sample_iter:
            size:
            raise_when_stopped:
        Returns:
            added: int, size of added samples.
        """
        size = size or self._max_size
        # Approximate
        init_size = self.effective_size
        while self.effective_size < size:
            try:
                self.push(next(sample_iter))
            except StopIteration as e:
                if raise_when_stopped:
                    raise e
                else:
                    break
        return self.effective_size - init_size

    def from_deque(self, sample_list: deque, size: int = None):
        size = size or self._max_size
        init_size = self.effective_size
        while self.effective_size < size:
            try:
                self.push(sample_list.popleft())
            except IndexError:
                break
        return self.effective_size - init_size

    def sort(self, key_fn: Callable[[Any], int]):
        if key_fn is not None:
            keys = numpy.array(list(map(key_fn, self.samples)))
            indices = numpy.argsort(keys)
            self._sort_idx = indices
            samples = list(self._samples)  # convert to list for faster indexing
            self._samples = deque([samples[i] for i in indices])

    def revert(self, samples: List[Any] = None):
        if self._sort_idx is None:
            return samples
        else:
            indices = numpy.argsort(self._sort_idx)
            indices = list(indices)

        arg_not_none=samples is not None

        if samples is None:
            samples=list(self._samples)
        else:
            if len(samples) != len(self._sort_idx):
                raise RuntimeError(
                    f'The number of samples ({len(samples)}) must match '
                    f'the number of indices ({len(self._sort_idx)}).'
                )
        samples=[samples[i] for i in indices]

        if arg_not_none:
            return samples
        else:
            self._samples=deque(samples)

    def process(self, collate_fn):
        self._data = collate_fn(self.samples)

    @overrides
    def state_dict(self) -> Dict:
        return get_state_dict(self, exclusions=self._exclusions)

    @overrides
    def load_state_dict(self, state_dict: Dict) -> None:
        load_state_dict(self, state_dict)


class Cache(Batch, Iterable):
    def __init__(self, max_size: int, size_fn: Callable[[Any], int] = None):
        super().__init__(max_size, size_fn)

    def __next__(self):
        if self._samples:
            return self.pop()
        raise StopIteration

    def __iter__(self):
        return self



from typing import Any

from overrides import overrides

from lunas.readers.base import BaseReader


class TextLine(BaseReader):
    def __init__(self, filename: str, bufsize: int = 10000, num_threads: int = 1):
        super().__init__(bufsize, num_threads)
        self._filename = filename
        self._fd = None
        self._iterator = None
        self._inclusions += ['_filename', '_num_line']
        self._num_line = -1

    @overrides
    def size(self) -> int:
        if self._num_line < 0:
            n = 0
            with open(self._filename) as r:
                for _ in r:
                    n += 1
            self._num_line = n
        return self._num_line

    @overrides
    def next(self) -> Any:
        line = next(self._iterator)
        return line

    @overrides
    def _reset_cursor(self):
        self._reset()
        super()._reset_cursor()

    def _reset(self):
        self._finalize()
        self._fd = open(self._filename)
        self._iterator = iter(self._fd)

    @overrides
    def _finalize(self):
        if self._fd is not None and not self._fd.closed:
            self._fd.close()
        self._iterator = None
        super()._finalize()
